#
#   Pyrex - Symbol Table
#

from Errors import error
import Naming
from PyrexTypes import \
  py_object_type, c_char_array_type, \
  CEnumType, CStructOrUnionType

class Entry:
  # name         string     Python name of entity
  # cname        string     C name of entity
  # type         PyrexType  type of entity
  # doc          string     docstring
  # borrowed     bool       is a borrowed reference
  # init         string     initial value
  # extern_flag  boolean    is declared extern
  # is_builtin   boolean    is a Python builtin name
  # is_pyglobal  boolean    is a Python module-level variable
  #                           or class attribute during
  #                           class construction
  # is_variable  boolean    is a variable
  # is_cfunction boolean    is a C function
  # is_type      boolean    is a type definition
  # is_const     boolean    is a constant
  # func_cname   string     C func implementing Python func
  # pos          position   source position where declared
  # namespace_cname string  if is_pyglobal, C variable
  #                           holding its home namespace
  # pymethdef_cname string  PyMethodDef structure
  
  borrowed = 0
  init = ""
  extern_flag = 0
  is_builtin = 0
  is_pyglobal = 0
  is_variable = 0
  is_cfunction = 0
  is_type = 0
  is_const = 0
  doc = None
  
  def __init__(self, name, cname, type, pos = None, init = None):
    self.name = name
    self.cname = cname
    self.type = type
    self.pos = pos
    self.init = init


class Scope:
  # outer_scope       Scope or None      Enclosing scope
  # entries           {string : Entry}   Python name to entry mapping
  # const_entries     [Entry]            Constant entries
  # sue_entries       [Entry]            Struct/union/enum entries
  # arg_entries       [Entry]            Function argument entries
  # var_entries       [Entry]            User-defined variable entries
  # pyfunc_entries    [Entry]            Python function entries
  # cfunc_entries     [Entry]            C function entries
  # temp_entries      [Entry]            Temporary variable entries
  # free_temp_entries [Entry]            Temp variables currently unused
  # temp_counter      integer            Counter for naming temp vars
  # cname_to_entry    {string : Entry}   Temp cname to entry mapping
  # pow_function_used boolean            The C pow() function is used
  # return_type       PyrexType or None  Return type of function owning scope
  # is_class_scope    boolean            Is a Python class scope
  # scope_prefix      string             Disambiguator for C names

  is_class_scope = 0
  scope_prefix = ""
  
  def __init__(self, outer_scope):
    self.outer_scope = outer_scope
    self.entries = {}
    self.const_entries = []
    self.sue_entries = []
    self.arg_entries = []
    self.var_entries = []
    self.pyfunc_entries = []
    self.cfunc_entries = []
    self.temp_entries = []
    self.free_temp_entries = []
    self.pending_temp_entries = [] # TEMPORARY
    self.temp_counter = 1
    self.cname_to_entry = {}
    self.pow_function_used = 0
  
  def global_scope(self):
    # Return the module-level scope containing this scope.
    return self.outer_scope.global_scope()
  
  def declare(self, name, cname, type, pos):
    # Create new entry and add to dictionary. Reports an
    # error if already declared.
    d = self.entries
    if d.has_key(name):
      error(pos, "'%s' redeclared" % name)
    else:
      entry = Entry(name, cname, type, pos = pos)
      d[name] = entry
      return entry
  
  def declare_const(self, name, type, value, pos):
    # Add an entry for a named constant.
    entry = self.declare(name, name, type, pos)
    entry.is_const = 1
    entry.value = value
    return entry
  
  def declare_type(self, name, type, pos):
    # Add an entry for a type definition.
    entry = self.declare(name, name, type, pos)
    entry.is_type = 1
    return entry
    
  def declare_struct_or_union(self, name, kind, scope, pos):
    # Add an entry for a struct or union definition.
    entry = self.lookup_here(name)
    if not entry:
      type = CStructOrUnionType(name, kind, scope)
      entry = self.declare_type(name, type, pos)
      self.sue_entries.append(entry)
    else:
      if (entry.is_type 
          and entry.type.is_struct_or_union
          and (not scope or not entry.type.scope)):
        entry.type.scope = scope
      else:
        error(pos, "'%s' already defined at this level" % name)
    return entry
  
  def declare_enum(self, name, pos):
    type = CEnumType(name)
    entry = self.declare_type(name, type, pos)
    self.sue_entries.append(entry)
    return entry	
  
  def declare_var(self, name, type, pos, extern_flag = 0):
    # Add an entry for a variable.
    if extern_flag:
      cname = name
    else:
      cname = Naming.var_prefix + name
    entry = self.declare(name, cname, type, pos)
    entry.is_variable = 1
    entry.extern_flag = extern_flag
    return entry
    
  def declare_builtin(self, name, pos):
    return self.outer_scope.declare_builtin(name, pos)
  
  def declare_pyfunction(self, name, pos):
    # Add an entry for a Python function.
    entry = self.declare_var(name, py_object_type, pos)
    entry.func_cname = \
      Naming.func_prefix + self.scope_prefix + name
    entry.doc_cname = \
      Naming.funcdoc_prefix + self.scope_prefix + name
    entry.pymethdef_cname = \
      Naming.pymethdef_prefix + self.scope_prefix + name
    self.pyfunc_entries.append(entry)
    return entry
  
  def declare_cfunction(self, name, type, pos, extern_flag):
    # Add an entry for a C function.
    if extern_flag:
      cname = name
    else:
      cname = Naming.func_prefix + name
    entry = self.declare(name, cname, type, pos)
    entry.func_cname = cname
    entry.is_cfunction = 1
    entry.extern_flag = extern_flag
    self.cfunc_entries.append(entry)
    return entry
  
  def find(self, name, pos):
    # Look up name, report error if not found.
    entry = self.lookup(name)
    if entry:
      return entry
    else:
      error(pos, "'%s' not defined" % name)
  
  def lookup(self, name):
    # Look up name in this scope or an enclosing one.
    # Return None if not found.
    return (self.lookup_here(name)
      or (self.outer_scope and self.outer_scope.lookup(name))
      or None)

  def lookup_here(self, name):
    # Look up in this scope only, return None if not found.
    return self.entries.get(name, None)
    
  def lookup_target(self, name):
    # Look up name in this scope only. Declare as Python
    # variable if not found.
    entry = self.lookup_here(name)
    if not entry:
      entry = self.declare_var(name, py_object_type, None)
    return entry
  
  def add_string_const(self, value):
    # Add an entry for a string constant.
    cname = self.new_const_cname()
    entry = Entry("", cname, c_char_array_type, init = value)
    self.const_entries.append(entry)
    return entry
  
  def allocate_temp(self, type):
    # Allocate a temporary variable of the given type from the 
    # free list if available, otherwise create a new one.
    # Returns the cname of the variable.
    for entry in self.free_temp_entries:
      if entry.type == type:
        self.free_temp_entries.remove(entry)
        return entry.cname
    n = self.temp_counter
    self.temp_counter = n + 1
    cname = "%s%d" % (Naming.pyrex_prefix, n)
    entry = Entry("", cname, type)
    if type.is_pyobject:
      entry.init = "0"
    self.cname_to_entry[entry.cname] = entry
    self.temp_entries.append(entry)
    return entry.cname
  
  def allocate_temp_pyobject(self):
    # Allocate a temporary PyObject variable.
    return self.allocate_temp(py_object_type)
  
  def release_temp(self, cname):
    # Release a temporary variable for re-use.
    ## TEMPORARY HACK until flaw in algorithm
    ## for temp allocation can be redressed:
    ## Don't really release any temps until
    ## a whole expression has been processed.
    self.pending_temp_entries.append(self.cname_to_entry[cname])
    return
    # Originally:
    self.free_temp_entries.append(self.cname_to_entry[cname])
  
  def recycle_pending_temps(self):
    self.free_temp_entries.extend(self.pending_temp_entries)
    self.pending_temp_entries = []
  
  def use_utility_code(self, new_code):
    self.global_scope().use_utility_code(new_code)
  
  def generate_library_function_declarations(self, code):
    # Generate extern decls for C library funcs used.
    if self.pow_function_used:
      code.putln("extern double pow(double, double);")
    

class LocalScope(Scope):
  
#	def lookup(self, name):
#		# Look up name in this scope or an enclosing one.
#		return self.lookup_here(name) or \
#			self._global_scope.lookup(name)
  
  def declare_arg(self, name, type, pos):
    # Add an entry for an argument of a function.
    entry = self.declare(name, Naming.var_prefix + name, type, pos)
    entry.is_variable = 1
    if type.is_pyobject:
      entry.init = "0"
    #entry.borrowed = 1 # Not using borrowed arg refs for now
    self.arg_entries.append(entry)
    return entry
  
  def declare_var(self, name, type, pos, extern_flag = 0):
    # Add an entry for a local variable.
    entry = Scope.declare_var(self, name, type, pos, extern_flag)
    self.var_entries.append(entry)
    return entry
  
  def declare_global(self, name, pos):
    # Pull entry from global scope into local scope.
    if self.lookup_here(name):
      error(pos, "'%s' redeclared")
    else:
      entry = self.global_scope().lookup_target(name)
      self.entries[name] = entry
    
  def new_const_cname(self):
    # Create a new globally-unique name for a constant.
    return self.global_scope().new_const_cname()


class ModuleScope(Scope):
  # module_name          string       Python name of the module
  # module_cname         string       C name of Python module object
  # #module_dict_cname    string       C name of module dict object
  # method_table_cname   string       C name of method table
  # doc                  string       Module doc string
  # doc_cname            string       C name of module doc string
  # const_counter        integer      Counter for naming constants
  # utility_code_used    [string]     Utility code to be included

  def __init__(self, name):
    Scope.__init__(self, BuiltinScope())
    self.module_name = name
    self.module_cname = Naming.module_cname
    self.module_dict_cname = Naming.moddict_cname
    self.method_table_cname = Naming.methtable_cname
    self.doc = ""
    self.doc_cname = Naming.moddoc_cname
    self.const_counter = 1
    self.utility_code_used = []
  
  def global_scope(self):
    return self
  
  def declare_var(self, name, type, pos, extern_flag = 0):
    # Add an entry for a global variable. If it is a Python
    # object type, it will live in the module dictionary,
    # otherwise it will be a C global variable.
    entry = Scope.declare_var(self, name, type, pos, extern_flag)
    if type.is_pyobject:
      entry.is_pyglobal = 1
      entry.namespace_cname = self.module_cname
    else:
      self.var_entries.append(entry)
    return entry
  
  def declare_global(self, name, pos):
    pass
    
  def new_const_cname(self):
    # Create a new globally-unique name for a constant.
    n = self.const_counter
    self.const_counter = n + 1
    return "%s%d" % (Naming.const_prefix, n)
  
  def use_utility_code(self, new_code):
    #  Add string to list of utility code to be included,
    #  if not already there (tested using 'is').
    for old_code in self.utility_code_used:
      if old_code is new_code:
        return
    self.utility_code_used.append(new_code)


class BuiltinScope(Scope):
  #  The builtin namespace.
  
  def __init__(self):
    Scope.__init__(self, None)
  
  def declare_builtin(self, name, pos):
    entry = self.declare(name, name, py_object_type, pos)
    entry.is_builtin = 1
    return entry
  

class StructOrUnionScope(Scope):
  #  Namespace of a C struct or union.

  def __init__(self):
    Scope.__init__(self, None)

  def declare_var(self, name, type, pos, extern_flag = 0):
    # Add an entry for an attribute.
    entry = self.declare(name, name, type, pos)
    entry.is_variable = 1
    self.var_entries.append(entry)
    if type.is_pyobject:
      error(pos,
        "C struct/union member cannot be a Python object")
    if extern_flag:
      error(pos,
        "C struct/union member cannot be declared extern")
    return entry


class ClassScope(Scope):
  #  Namespace of a Python class.
  #
  #  class_name        string  Name of the class
  #  class_dict_cname  string  C variable holding class dict
  #  class_obj_cname   string  C variable holding class object
  
  is_class_scope = 1
  
  def __init__(self, outer_scope, name):
    Scope.__init__(self, outer_scope)
    self.class_name = name
    self.scope_prefix = "%d%s_" % (len(name), name)
  
  def declare_var(self, name, type, pos, extern_flag = 0):
    # Add an entry for a class attribute.
    entry = Scope.declare_var(self, name, type, pos, extern_flag)
    entry.is_pyglobal = 1
    entry.namespace_cname = self.class_obj_cname
    return entry

  def allocate_temp(self, type):
    return self.outer_scope.allocate_temp(type)

  def release_temp(self, cname):
    self.outer_scope.release_temp(cname)

  def recycle_pending_temps(self):
    self.outer_scope.recycle_pending_temps()
